"""The implementation of the physics of the simulated environment."""

from typing import Optional, Any, Dict, Iterator, Tuple

from shrdlu_blocks.geometry import Point
from shrdlu_blocks.scenes import Scene, PhysicalObject
from shrdlu_blocks.typedefs import UnmetConditionError, ObjectID


__all__ = ['Controller']


class Controller:
    """The controller, through which all queries and commands between the AI
    and the simulated environment must pass."""

    # IMPORTANT: Public methods must *never* accept or return mutable Scene components. The entire
    #            point of this class is to ensure that the information and control flows between the
    #            environment and AI controller are constrained. If the simulated environment is
    #            unconstrained, it doesn't provide a useful test environment for the AI controller.

    def __init__(self, scene: Scene, grasper_distance_tolerance: float = 0.000001):
        self._scene = scene
        self._grasper_distance_tolerance = grasper_distance_tolerance
        self._default_grasper: Optional[PhysicalObject] = None
        self._grasper_cache: Dict[ObjectID, PhysicalObject] = {}

    @property
    def default_grasper(self) -> Optional[ObjectID]:
        """The default grasper that is used when no grasper is specified in
        calls to methods which require one."""
        if self._default_grasper is None:
            return None
        return self._default_grasper.tags.get('obj_id', None)

    @default_grasper.setter
    def default_grasper(self, grasper_id: ObjectID) -> None:
        """The default grasper that is used when no grasper is specified in
        calls to methods which require one."""
        self._default_grasper = self._require_grasper(grasper_id)

    def grasper_is_closed(self, grasper_id: ObjectID = None) -> bool:
        """Query whether the grasper is closed. Returns a boolean value."""
        return self._require_grasper(grasper_id).tags.get('closed', False)

    def grasper_is_lowered(self, grasper_id: ObjectID = None) -> bool:
        """Query whether the grasper is lowered. Returns a boolean value."""
        return self._require_grasper(grasper_id).tags.get('lowered', False)

    def get_grasped_object(self, grasper_id: ObjectID = None) -> Optional[ObjectID]:
        """Query which object is currently grasped by the grasper. Returns the
        object ID of the grasped object, or None if no object is currently
        grasped."""
        return self._require_grasper(grasper_id).tags.get('grasped', None)

    def close_grasper(self, grasper_id: ObjectID = None) -> None:
        """Attempt to close the grasper. If the grasper is resting on a
        graspable object and is situated at its center of mass, the grasper
        will begin grasping the object."""
        grasper = self._require_grasper(grasper_id)
        del grasper_id
        if grasper.tags.get('closed', False):
            raise UnmetConditionError('Grasper is not open.')
        if grasper.tags.get('lowered', False):
            # Find out if the grasper was lowered onto a graspable object.
            # If so, make it grasp the object.
            resting_on_id = grasper.tags.get('resting_on', None)
            if resting_on_id:
                resting_on = self._require_specific_object(resting_on_id)
                if resting_on.tags.get('graspable', False):
                    grasper.tags['grasped'] = resting_on_id
                    resting_on.tags['grasped_by'] = grasper.tags.get('obj_id', None)
                    target, target_height = self._find_object_below_grasper(grasper)
                    if target:
                        grasper.tags['resting_on'] = target.tags.get('obj_id', None)
        grasper.tags['closed'] = True

    def open_grasper(self, grasper_id: ObjectID = None) -> None:
        """Attempt to open the grasper. If the grasper is holding an object and
        the object is supported, the object will be released."""
        grasper = self._require_grasper(grasper_id)
        del grasper_id
        if not grasper.tags.get('closed', False):
            raise UnmetConditionError('Grasper is not closed.')
        grasped_obj_id = grasper.tags.get('grasped', None)
        if grasped_obj_id is not None:
            grasped_obj = self._require_specific_object(grasped_obj_id)
            if not grasper.tags.get('lowered', False):
                raise UnmetConditionError('Grasper must be lowered first when holding an object.')
            # If an object was grasped, we have to be above another object that can support it.
            resting_on_id = grasper.tags.get('resting_on', None)
            if resting_on_id is None:
                raise UnmetConditionError('Object must be lowered onto another object that can '
                                          'support it in order to be dropped.')
            resting_on = self._require_specific_object(resting_on_id)
            if not resting_on.can_support(grasped_obj):
                raise UnmetConditionError('Object must be lowered onto another object that can '
                                          'support it in order to be dropped.')
            assert grasped_obj.tags.get('grasped_by', None) == grasper.tags.get('obj_id', None)
            # Let go of the object.
            grasped_obj.tags['grasped_by'] = None
            grasper.tags['grasped'] = None
            grasper.tags['resting_on'] = grasped_obj_id
        grasper.tags['closed'] = False

    def move_grasper(self, x: float = None, y: float = None, grasper_id: ObjectID = None) -> None:
        """Attempt to move the grasper to a new (x, y) coordinate. One or both
        coordinates must be specified. If either coordinate is omitted, its
        current value will be assumed."""
        if x is not None and not isinstance(x, (int, float)):
            raise TypeError(x)
        if y is not None and not isinstance(y, (int, float)):
            raise TypeError(y)
        grasper = self._require_grasper(grasper_id)
        del grasper_id
        if grasper.tags.get('lowered', False):
            raise UnmetConditionError('Grasper must be raised before it can be moved.')
        if x is None and y is None:
            raise UnmetConditionError('No position specified.')
        if x is not None and not grasper.tags.get('min_x', x) <= x <= grasper.tags.get('max_x', x):
            raise UnmetConditionError('The grasper cannot move there.')
        if y is not None and not grasper.tags.get('min_y', y) <= y <= grasper.tags.get('max_y', y):
            raise UnmetConditionError('The grasper cannot move there.')
        old_x, old_y, old_z = grasper.position
        grasper.position = Point(old_x if x is None else x, old_y if y is None else y, old_z)
        grasped_obj_id = grasper.tags.get('grasped', None)
        if grasped_obj_id is not None:
            grasped_obj = self._require_specific_object(grasped_obj_id)
            highest_point = grasped_obj.find_highest_point()
            height = highest_point.z - grasped_obj.position.z
            grasped_obj.position = grasper.position - Point(0, 0, height)

    def lower_grasper(self, grasper_id: ObjectID = None) -> None:
        """Attempt to lower the grasper. The grasper will be lowered until it
        or the object it is holding is resting on an object below it, or until
        the grasper is maximally extended, whichever comes first. (The grasper
        can be extended down to 'ground level', where the z coordinate is zero,
        and no further.)"""
        grasper = self._require_grasper(grasper_id)
        del grasper_id
        if grasper.tags.get('lowered', False):
            raise UnmetConditionError('Grasper is not raised.')
        target, target_height = self._find_object_below_grasper(grasper)
        grasped_obj_id = grasper.tags.get('grasped', None)
        if grasped_obj_id is None:
            # Set the grasper's position so that the grasper is resting on whatever object is below
            # it.
            grasper.position = Point(grasper.position.x,
                                     grasper.position.y,
                                     target_height)
            if target:
                x_displacement, y_displacement = (target.position - grasper.position)[:2]
                distance = (x_displacement ** 2 + y_displacement ** 2) ** 0.5
                if distance <= self._grasper_distance_tolerance:
                    grasper.tags['resting_on'] = target.tags.get('obj_id', None)
        else:
            # Set the grasper's position so that the grasped object is resting on whatever object is
            # below it. Remember to update both the grasper's and the object's positions.
            grasped_obj = self._require_specific_object(grasped_obj_id)
            highest_point = grasped_obj.find_highest_point()
            height = highest_point.z - grasped_obj.position.z
            grasper.position = Point(grasper.position.x, grasper.position.y, target_height + height)
            grasped_obj.position = Point(grasper.position.x, grasper.position.y, target_height)
            if target:
                grasper.tags['resting_on'] = target.tags.get('obj_id', None)
        grasper.tags['lowered'] = True

    def raise_grasper(self, grasper_id: ObjectID = None) -> None:
        """Attempt to raise the grasper. The grasper will be raised until it
        and the object it is holding, if any, are high enough to safely clear
        all objects in the scene."""
        grasper = self._require_grasper(grasper_id)
        del grasper_id
        if not grasper.tags.get('lowered', False):
            raise UnmetConditionError('Grasper is not lowered.')
        minimum_height = self._find_highest_stable_point() + 0.1
        grasped_obj_id = grasper.tags.get('grasped', None)
        if grasped_obj_id is None:
            grasper.position = Point(grasper.position.x, grasper.position.y, minimum_height)
        else:
            grasped_obj = self._require_specific_object(grasped_obj_id)
            grasped_obj.position = Point(grasper.position.x, grasper.position.y, minimum_height)
            highest_point = grasped_obj.find_highest_point()
            grasper.position = Point(grasper.position.x, grasper.position.y, highest_point.z)
        grasper.tags['lowered'] = False
        grasper.tags['resting_on'] = None

    def find_objects(self, **tags) -> Iterator[ObjectID]:
        """Query the objects in the scene. An iterator will be returned over
        all objects in the scene whose tag values match the given keyword
        arguments exactly. If no keyword arguments are provided, an iterator
        over all objects in the scene will be returned."""
        for obj in self._scene.find_objects(**tags):
            obj_id = obj.tags.get('obj_id', None)
            if obj_id is not None:
                yield obj_id

    def get_object_tag(self, obj_id: ObjectID, tag: str, default: Any = None) -> Any:
        """Query the value of an object's tag. If the object does not have a
        value for the given tag, the default is returned."""
        return self._require_specific_object(obj_id).tags.get(tag, default)

    def iter_object_tags(self, obj_id: ObjectID) -> Iterator[Tuple[str, Any]]:
        """Query the tag key/value pairs associated with the object."""
        yield  from self._require_specific_object(obj_id).tags.items()

    def get_object_position(self, obj_id: ObjectID) -> Point:
        """Query the position of an object in the scene."""
        return self._require_specific_object(obj_id).position

    def _get_specific_grasper(self, grasper_id: ObjectID) -> Optional[PhysicalObject]:
        """Return the specific grasper indicated by the given grasper ID. If no
        such grasper exists, return None."""
        if not isinstance(grasper_id, int):
            raise TypeError(grasper_id)
        if grasper_id in self._grasper_cache:
            return self._grasper_cache[grasper_id]
        grasper = self._get_specific_object(grasper_id)
        if grasper is None:
            return None
        if grasper.tags.get('kind', None) != 'grasper':
            raise UnmetConditionError("Object is not a grasper.")
        self._grasper_cache[grasper_id] = grasper
        if self._default_grasper is None:
            self._default_grasper = grasper
        return grasper

    def _get_grasper(self, grasper_id: ObjectID = None) -> Optional[PhysicalObject]:
        """If a grasper ID is provided, return that specific grasper.
        Otherwise, return the default grasper. If the default grasper is not
        set, set it to an arbitrary grasper first. If no graspers exist in the
        scene, or the specific grasper requested does not exist, return None."""
        if grasper_id is not None:
            if not isinstance(grasper_id, int):
                raise TypeError(grasper_id)
            return self._get_specific_grasper(grasper_id)
        if self._default_grasper is not None:
            return self._default_grasper
        for grasper in self._scene.find_objects(kind='grasper'):
            self._default_grasper = grasper
            return grasper
        return None

    def _require_grasper(self, grasper_id: ObjectID = None) -> PhysicalObject:
        """If a grasper ID is provided, return that specific grasper.
        Otherwise, return the default grasper. If the default grasper is not
        set, set it to an arbitrary grasper first. If no graspers exist in the
        scene, or the specific grasper requested does not exist, raise an
        exception."""
        if grasper_id is not None and not isinstance(grasper_id, int):
            raise TypeError(grasper_id)
        grasper = self._get_grasper(grasper_id)
        if grasper is None:
            raise UnmetConditionError('Grasper not found.')
        return grasper

    def _get_specific_object(self, obj_id: ObjectID) -> Optional[PhysicalObject]:
        """Return the specific object indicated by the given object ID. If no
        such object exists, return None."""
        if not isinstance(obj_id, int):
            raise TypeError(obj_id)
        for obj in self._scene.find_objects(obj_id=obj_id):
            return obj
        return None

    def _require_specific_object(self, obj_id: ObjectID) -> PhysicalObject:
        """Return the specific object indicated by the given object ID. If no
        such object exists, raise an exception."""
        if not isinstance(obj_id, int):
            raise TypeError(obj_id)
        obj = self._get_specific_object(obj_id)
        if obj is None:
            raise UnmetConditionError('Object not found.')
        return obj

    def _find_object_below_grasper(self, grasper: PhysicalObject) \
            -> Tuple[Optional[PhysicalObject], float]:
        """Find the highest object directly below the grasper. Return a tuple
        of the form (obj, dist), where obj is the object identified and dist is
        the vertical distance from the grasper to the highest point of the
        object. If no object is found below the grasper, return (None, 0)."""
        if not isinstance(grasper, PhysicalObject):
            raise TypeError(grasper)
        # Find the object the grasper is above, if any, and its height above ground level.
        target = None
        target_height = 0
        for obj in self._scene.objects:
            if not (obj.tags.get('grasped_by', None) is None and
                    obj.is_below_point(grasper.position)):
                continue
            obj_height = obj.find_highest_point().z
            if target_height <= obj_height:
                target = obj
                target_height = obj_height
        return target, target_height

    def _find_highest_stable_point(self) -> float:
        """Find the highest point of the highest stably positioned object. An
        object is considered stably positioned if it is either immovable or
        resting on another stably positioned object."""
        result = 0
        for obj in self._scene.objects:
            if obj.tags.get('kind') == 'grasper' or obj.tags.get('grasped_by', None) is not None:
                continue
            obj_result = obj.find_highest_point()
            if obj_result is not None and obj_result.z > result:
                result = obj_result.z
        return result



"""Simple type definitions with no dependencies."""

from typing import NamedTuple


__all__ = ['UnmetConditionError', 'ObjectID', 'Color']


class UnmetConditionError(Exception):
    """A required condition for a command or query issued to a controller was
    not met."""


class ObjectID(int):
    """A unique identifier for each object in a scene."""


# An RGB-formatted color specifier.
Color = NamedTuple('Color', [('red', int), ('green', int), ('blue', int)])



"""Functionality pertaining to the arrangement and state of objects within the
simulated environment."""

import time
from dataclasses import dataclass
from typing import Tuple, Dict, Any, Optional, Iterator

from shrdlu_blocks.geometry import PolygonalShape, Point, inside_triangle, make_grasper, \
    make_table, make_block, make_pyramid, make_box
from shrdlu_blocks.typedefs import ObjectID, Color


__all__ = ['PhysicalObject', 'Scene', 'make_standard_scene']


@dataclass
class PhysicalObject:
    """A physical object that can appear within a scene."""

    # The shape of the object. Points in the shape are positioned relative to
    # the shape's center of mass.
    shape: PolygonalShape

    # The color of the object.
    color: Color

    # The position of the object's center of mass within the scene.
    position: Point

    # Metadata attached to the object.
    tags: Dict[str, Any]

    def __str__(self) -> str:
        kind = 'object'
        obj_id = None
        attributes = []
        for key, value in sorted(self.tags.items()):
            if key == 'kind':
                kind = value
            elif key == 'obj_id':
                obj_id = value
            elif isinstance(value, bool):
                if value:
                    attributes.append(key)
            elif isinstance(value, str):
                if value not in attributes:
                    attributes.append(value)
            elif value is not None:
                attributes.append('%s=%r' % (key, value))
        if obj_id is None:
            return kind + '[' + ','.join(attributes) + ']'
        else:
            return kind + '#' + str(obj_id) + '[' + ','.join(attributes) + ']'

    def find_highest_point(self) -> Optional[Point]:
        """Return the highest salient point of the object, or None if the
        object contains no salient points."""
        if self.tags.get('kind', None) == 'box':
            # Boxes get special treatment. We ignore their sides.
            highest_point = self.position
        else:
            highest_point = max(self.shape.iter_salient_points(), key=lambda p: p.z, default=None)
        if highest_point is None:
            return None
        return self.position + highest_point

    def is_below_point(self, point: Point) -> bool:
        """Return whether any part of the object is directly underneath the
        given point."""
        relative_position = point - self.position
        point_projection = Point(relative_position.x, relative_position.y, 0)
        for surface in self.shape.surfaces:
            surface_projection = sorted({Point(p.x, p.y, 0) for p in surface.iter_salient_points()},
                                        key=lambda p: tuple(p))
            # Find any triangle that contains the point.
            for index1, a in enumerate(surface_projection):
                for index2 in range(index1 + 1, len(surface_projection)):
                    b = surface_projection[index2]
                    for index3 in range(index2 + 1, len(surface_projection)):
                        c = surface_projection[index3]
                        if inside_triangle(point_projection, (a, b, c)):
                            return True
        return False

    def can_support(self, obj: 'PhysicalObject') -> bool:
        """Return whether this object can support the other given their current
        (x, y) coordinates. (The relative elevation, z, is ignored.)"""
        if not self.tags.get('can_support', False):
            return False
        kind = self.tags.get('kind', None)
        if kind == 'box':
            # Boxes get special treatment; we cannot set something down if it would land on the side
            # of the box.
            return all(self.is_below_point(obj.position + point)
                       for point in obj.shape.iter_salient_points())
        return self.is_below_point(obj.position)


@dataclass
class Scene:
    """An arrangement of objects within a larger space."""

    # The objects appearing in the scene.
    objects: Tuple[PhysicalObject, ...]

    # Metadata attached to the scene.
    tags: Dict[str, Any]

    def find_objects(self, **tags) -> Iterator[PhysicalObject]:
        """Return an iterator over all objects whose tag values precisely match
        the provided keyword arguments. If no keyword arguments are provided,
        an iterator over all objects is returned."""
        for obj in self.objects:
            if all(obj.tags.get(key, None) == value for key, value in tags.items()):
                yield obj


def make_standard_scene() -> Scene:
    """Construct the standard scene and return it. The standard scene is an
    approximation of the initial arrangement of objects as appearing in the
    original SHDRLU program."""

    grasper = PhysicalObject(
        make_grasper(0.05, 1),
        Color(230, 230, 0),
        Point(0, 0, 0.5),
        dict(kind='grasper', color='yellow', graspable=False, can_support=False)
    )

    table = PhysicalObject(
        make_table(1, 1),
        Color(0, 0, 0),
        Point(0, 0, 0),
        dict(kind='table', color='black', graspable=False, can_support=True)
    )

    big_red_block = PhysicalObject(
        make_block(0.15, 0.15, 0.15),
        Color(255, 0, 0),
        Point(-0.3, 0.1, 0),
        dict(kind='block', size='big', height='tall', width='wide', color='red', graspable=True,
             can_support=True)
    )

    small_red_block = PhysicalObject(
        make_block(0.1, 0.1, 0.08),
        Color(255, 0, 0),
        Point(-0.25, -0.2, 0),
        dict(kind='block', size='small', height='short', width='narrow', color='red',
             graspable=True, can_support=True)
    )

    small_green_pyramid = PhysicalObject(
        make_pyramid(0.1, 0.1, 0.08),
        Color(0, 255, 0),
        Point(-0.25, -0.2, 0.08),
        dict(kind='pyramid', size='small', height='short', width='narrow', color='green',
             graspable=True, can_support=False)
    )

    medium_sized_green_block = PhysicalObject(
        make_block(0.15, 0.15, 0.1),
        Color(0, 255, 0),
        Point(-0.3, 0.05, 0.15),
        dict(kind='block', size='medium', height='medium', width='wide', color='green',
             graspable=True, can_support=True)
    )

    tall_blue_block = PhysicalObject(
        make_block(0.15, 0.15, 0.2),
        Color(0, 0, 255),
        Point(-0.1, 0.4, 0),
        dict(kind='block', size='big', height='tall', width='medium', color='blue', graspable=True,
             can_support=True)
    )

    big_green_block = PhysicalObject(
        make_block(0.2, 0.2, 0.15),
        Color(0, 255, 0),
        Point(0.1, -0.15, 0),
        dict(kind='block', size='big', height='tall', width='wide', color='green', graspable=True,
             can_support=True)
    )

    tall_red_pyramid = PhysicalObject(
        make_pyramid(0.1, 0.1, 0.2),
        Color(255, 0, 0),
        Point(0.15, -0.1, 0.15),
        dict(kind='pyramid', size='medium', height='tall', width='narrow', color='red',
             graspable=True, can_support=False)
    )

    big_white_box = PhysicalObject(
        make_box(0.35, 0.35, 0.2),
        Color(255, 255, 255),
        Point(0.25, 0.25, 0),
        dict(kind='box', size='big', height='medium', width='wide', color='white', graspable=False,
             can_support=True)
    )

    wide_blue_pyramid = PhysicalObject(
        make_pyramid(0.15, 0.15, 0.1),
        Color(0, 0, 255),
        Point(0.25, 0.25, 0),
        dict(kind='pyramid', size='medium', height='medium', width='medium', color='blue',
             graspable=True, can_support=False)
    )

    objects = [
        grasper,
        table,
        big_red_block,
        small_red_block,
        small_green_pyramid,
        medium_sized_green_block,
        tall_blue_block,
        big_green_block,
        tall_red_pyramid,
        big_white_box,
        wide_blue_pyramid,
    ]
    for index, obj in enumerate(objects):
        obj.tags['obj_id'] = ObjectID(index)
    return Scene(tuple(objects), dict(created=time.time()))



"""
A simple demo of the environment.

Usage:
    python3 -m shrdlu_blocks.demo

The environment will be displayed in a graphics window. The user can type
various commands into the graphics window to query the scene and control the
grasper. Type `help` to get a list of commands.
"""

import ast
import io
import sys
import traceback

import pygame.display

from shrdlu_blocks.control import Controller
from shrdlu_blocks.typedefs import UnmetConditionError, ObjectID
from shrdlu_blocks.scenes import make_standard_scene, PhysicalObject
from shrdlu_blocks.viewer import Viewer


__all__ = ['demo']


def demo_callback(controller: Controller, command: str) -> str:
    """Parse and execute the command."""
    if command == 'exit':
        pygame.quit()
        sys.exit(0)
    output_buffer = io.StringIO()
    if command == 'help':
        print("Commands:", file=output_buffer)
        print("    help", file=output_buffer)
        print("    exit", file=output_buffer)
        for name in dir(controller):
            if not name.startswith('_'):
                print("    " + name, file=output_buffer)
        return output_buffer.getvalue()
    pieces = command.split()
    command = pieces.pop(0)
    if not command:
        return output_buffer.getvalue()
    if '.' in command or command.startswith('_') or command not in dir(controller):
        print("ERROR: Invalid command", file=output_buffer)
        return output_buffer.getvalue()
    # noinspection PyBroadException
    try:
        args = []
        for piece in pieces:
            try:
                arg = ast.literal_eval(piece)
            except ValueError:
                arg = piece
            args.append(arg)
        result = getattr(controller, command)(*args)
        if result is None:
            pass
        elif isinstance(result, str) or not hasattr(result, '__iter__'):
            print(repr(result), file=output_buffer)
        else:
            object_count = len(list(controller.find_objects()))
            for item in result:
                if ('objects' in command and isinstance(item, int) and
                        0 <= item < object_count):
                    tags = dict(controller.iter_object_tags(ObjectID(item)))
                    # Cheat just a little by constructing a mock object with the tags so we can use
                    # the __str__() method it defines.
                    # noinspection PyTypeChecker
                    mock_obj = PhysicalObject(None, None, None, tags)
                    print(str(mock_obj), file=output_buffer)
                else:
                    print(repr(item), file=output_buffer)
    except UnmetConditionError as e:
        print(e, file=output_buffer)
    except Exception:
        traceback.print_exc(file=output_buffer)
    return output_buffer.getvalue()


def demo():
    """
    Let the user play around with the standard scene using programmatic
    instructions passed directly to the controller.

    The environment will be displayed in a graphics window. The user can type
    various commands into the graphics window to query the scene and control
    the grasper. Type `help` to get a list of commands.
    """

    pygame.init()

    screen_info = pygame.display.Info()
    screen_width = screen_info.current_w
    screen_height = screen_info.current_h
    screen = pygame.display.set_mode((screen_width // 2, screen_height // 2))

    Viewer(screen, "SHRDLU Blocks Demo", demo_callback).run()


if __name__ == '__main__':
    demo()



"""Functionality pertaining to the geometry of objects."""

import math
from abc import ABC, abstractmethod
from typing import Tuple, TypeVar, Generic, Callable, Type, Iterator, Iterable, Union


__all__ = ['GeometricObject', 'Point', 'Edge', 'PolygonalSurface', 'PolygonalShape',
           'make_table', 'make_box', 'make_block', 'make_pyramid', 'make_grasper',
           'triangle_area', 'inside_triangle']


Element = TypeVar('Element')
Self = TypeVar('Self')


class GeometricObject(Generic[Element], ABC):
    """Base class for geometric objects -- points, edges, surfaces, and shapes.

    Geometric objects are immutable, hashable, comparable, iterable, and
    indexable."""

    @classmethod
    @abstractmethod
    def _from_elements(cls: Type[Self], elements: Iterable[Element]) -> Self:
        raise NotImplementedError()

    @abstractmethod
    def _get_elements(self) -> Tuple[Element, ...]:
        raise NotImplementedError()

    def apply_elementwise_transform(self: Self, transform: Callable[[Element], Element]) -> Self:
        """Construct a new object of the same type by applying the same
        transform individually to each element."""
        return self._from_elements(transform(element) for element in self._get_elements())

    def iter_salient_points(self: Self) -> Iterator['Point']:
        """Iterate over the salient points in the object. Salient points are
        points that are explicitly mentioned in the object's structure, e.g.,
        the start and end of an edge."""
        for element in self._get_elements():
            yield from element.iter_salient_points()

    def __repr__(self) -> str:
        return type(self).__name__ + repr(self._get_elements())

    def __hash__(self) -> int:
        return hash(self._get_elements())

    def __eq__(self, other: 'GeometricObject') -> bool:
        if type(self) != type(other):
            return NotImplemented
        return self._get_elements() == other._get_elements()

    def __ne__(self, other: 'GeometricObject') -> bool:
        return not self == other

    def __iter__(self) -> Iterator[Element]:
        return iter(self._get_elements())

    def __len__(self) -> int:
        return len(self._get_elements())

    def __contains__(self, item: Element) -> bool:
        return item in self._get_elements()

    def __getitem__(self, index: int) -> Element:
        return self._get_elements()[index]

    def __add__(self: Self, other: 'Point') -> Self:
        return self.apply_elementwise_transform(lambda e: e + other)

    def __radd__(self: Self, other: 'Point') -> Self:
        return self.__add__(other)

    def __sub__(self: Self, other: 'Point') -> Self:
        return self.apply_elementwise_transform(lambda e: e - other)

    def __rsub__(self: Self, other: 'Point') -> Self:
        return self.apply_elementwise_transform(lambda e: other - e)

    def __mul__(self: Self, other: float) -> Self:
        return self.apply_elementwise_transform(lambda e: e * other)

    def __rmul__(self: Self, other: float) -> Self:
        return self.__mul__(other)

    def __truediv__(self: Self, other: float) -> Self:
        return self.apply_elementwise_transform(lambda e: e / other)

    def scale(self: Self, factor: float, center: 'Point' = None) -> Self:
        """Scale the object by the given factor. If center is provided, it is
        treated as the point whose position is invariant under the scaling
        transformation. If center is not provided, the origin is assumed."""
        if center:
            return center + factor * (self - center)
        else:
            return factor * self


class Point(GeometricObject[float]):
    """An arbitrary point in 3D space."""

    def __init__(self, x: float, y: float, z: float):
        self._x = x
        self._y = y
        self._z = z
        self._elements = (x, y, z)

    @property
    def x(self) -> float:
        """The coordinate of the point along the x dimension, which is negative
        to the left and positive to the right."""
        return self._x

    @property
    def y(self) -> float:
        """The coordinate of the point along the y dimension, which is positive
        in the direction the viewer is facing and negative behind the viewer."""
        return self._y

    @property
    def z(self) -> float:
        """The coordinate of the point along the z dimension, which is positive
        above 'ground level' and negative below it."""
        return self._z

    @classmethod
    def _from_elements(cls: Type[Self], elements: Iterable[float]) -> Self:
        return Point(*elements)

    def _get_elements(self) -> Tuple[float, ...]:
        return self._elements

    def iter_salient_points(self: Self) -> Iterator['Point']:
        """Iterate over the salient points in the object. Salient points are
        points that are explicitly mentioned in the object's structure, e.g.,
        the start and end of an edge."""
        yield self

    def __neg__(self) -> 'Point':
        return Point(-self.x, -self.y, -self.z)

    def __abs__(self) -> float:
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def __add__(self, other: 'Point') -> 'Point':
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: 'Point') -> 'Point':
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: Union[float, 'Point']) -> 'Point':
        if isinstance(other, (int, float)):
            return Point(self.x * other, self.y * other, self.z * other)
        elif isinstance(other, Point):
            return Point(self.x * other.x, self.y * other.y, self.z * other.z)
        else:
            return NotImplemented

    def __truediv__(self, other: Union[float, 'Point']) -> 'Point':
        if isinstance(other, (int, float)):
            return Point(self.x / other, self.y / other, self.z / other)
        elif isinstance(other, Point):
            return Point(self.x / other.x, self.y / other.y, self.z / other.z)
        else:
            return NotImplemented


class Edge(GeometricObject[Point]):
    """An arbitrary line segment in 3D space."""

    def __init__(self, start: Point, end: Point):
        self._start = start
        self._end = end
        self._elements = (start, end)

    @property
    def start(self) -> Point:
        """The point where the edge starts."""
        return self._start

    @property
    def end(self) -> Point:
        """The point where the edge ends."""
        return self._end

    @classmethod
    def _from_elements(cls: Type[Self], elements: Iterable[Point]) -> Self:
        return Edge(*elements)

    def _get_elements(self) -> Tuple[Point, Point]:
        return self._elements


class PolygonalSurface(GeometricObject[Edge]):
    """An arbitrary 2D polygon in 3D space.

    NOTE: The coplanarity of the points is assumed rather than validated by
          this class."""

    def __init__(self, edges: Tuple[Edge, ...]):
        self._edges = edges

    def __repr__(self) -> str:
        return type(self).__name__ + '(' + repr(self._edges) + ')'

    @property
    def edges(self) -> Tuple[Edge, ...]:
        """The edges of this surface."""
        return self._edges

    @classmethod
    def _from_elements(cls: Type[Self], elements: Iterable[Edge]) -> Self:
        return PolygonalSurface(tuple(elements))

    def _get_elements(self) -> Tuple[Edge, ...]:
        return self._edges


class PolygonalShape(GeometricObject[PolygonalSurface]):
    """An arbitrary 3D polygonal shape in 3D space.

    NOTE: Self-consistency is assumed rather than validated by this class."""

    def __init__(self, surfaces: Tuple[PolygonalSurface, ...]):
        self._surfaces = surfaces

    def __repr__(self) -> str:
        return type(self).__name__ + '(' + repr(self._surfaces) + ')'

    @property
    def surfaces(self) -> Tuple[PolygonalSurface, ...]:
        """The surfaces of this shape."""
        return self._surfaces

    @classmethod
    def _from_elements(cls: Type[Self], elements: Iterable[PolygonalSurface]) -> Self:
        return PolygonalShape(tuple(elements))

    def _get_elements(self) -> Tuple[PolygonalSurface, ...]:
        return self._surfaces


def make_block(width: float, depth: float, height: float) -> PolygonalShape:
    """Construct a PolygonalShape representing a box (a cuboid/rectangular
    prism) and return it."""
    left = -width / 2
    right = width / 2
    front = -depth / 2
    back = depth / 2
    top = height
    bottom = 0
    x_span = (left, right)
    y_span = (front, back)
    z_span = (bottom, top)
    corners = []
    for x in x_span:
        for y in y_span:
            for z in z_span:
                corners.append(Point(x, y, z))
    assert len(corners) == 8
    edges = []
    for index, point1 in enumerate(corners):
        for point2 in corners[index + 1:]:
            adjacent = sum(coord1 != coord2 for coord1, coord2 in zip(point1, point2)) == 1
            if adjacent:
                edges.append(Edge(point1, point2))
    assert len(edges) == 12
    surfaces = []
    for x in x_span:
        selected_edges = tuple(edge for edge in edges if edge.start[0] == edge.end[0] == x)
        assert len(selected_edges) == 4, (x, selected_edges)
        surface = PolygonalSurface(selected_edges)
        surfaces.append(surface)
    for y in y_span:
        selected_edges = tuple(edge for edge in edges if edge.start[1] == edge.end[1] == y)
        assert len(selected_edges) == 4
        surface = PolygonalSurface(selected_edges)
        surfaces.append(surface)
    for z in z_span:
        selected_edges = tuple(edge for edge in edges if edge.start[2] == edge.end[2] == z)
        assert len(selected_edges) == 4
        surface = PolygonalSurface(selected_edges)
        surfaces.append(surface)
    assert len(surfaces) == 6
    return PolygonalShape(tuple(surfaces))


def make_box(width: float, depth: float, height: float) -> PolygonalShape:
    """Construct a PolygonalShape representing an open-topped box (a
    cuboid/rectangular prism minus the top surface) and return it."""
    left = -width / 2
    right = width / 2
    front = -depth / 2
    back = depth / 2
    top = height
    bottom = 0
    x_span = (left, right)
    y_span = (front, back)
    z_span = (bottom, top)
    corners = []
    for x in x_span:
        for y in y_span:
            for z in z_span:
                corners.append(Point(x, y, z))
    assert len(corners) == 8
    edges = []
    for index, point1 in enumerate(corners):
        for point2 in corners[index + 1:]:
            adjacent = sum(coord1 != coord2 for coord1, coord2 in zip(point1, point2)) == 1
            if adjacent:
                edges.append(Edge(point1, point2))
    assert len(edges) == 12
    surfaces = []
    for x in x_span:
        selected_edges = tuple(edge for edge in edges if edge.start[0] == edge.end[0] == x)
        assert len(selected_edges) == 4
        surface = PolygonalSurface(selected_edges)
        surfaces.append(surface)
    for y in y_span:
        selected_edges = tuple(edge for edge in edges if edge.start[1] == edge.end[1] == y)
        assert len(selected_edges) == 4
        surface = PolygonalSurface(selected_edges)
        surfaces.append(surface)
    selected_edges = tuple(edge for edge in edges if edge.start[2] == edge.end[2] == bottom)
    assert len(selected_edges) == 4
    surface = PolygonalSurface(selected_edges)
    surfaces.append(surface)
    assert len(surfaces) == 5
    return PolygonalShape(tuple(surfaces))


def make_pyramid(width: float, depth: float, height: float) -> PolygonalShape:
    """Construct a PolygonalShape representing a rectangular pyramid and return
    it."""
    left = -width / 2
    right = width / 2
    front = -depth / 2
    back = depth / 2
    top = height
    bottom = 0
    x_span = (left, right)
    y_span = (front, back)
    bottom_corners = []
    for x in x_span:
        for y in y_span:
            bottom_corners.append(Point(x, y, bottom))
    assert len(bottom_corners) == 4
    peak = Point(sum(x_span) / 2, sum(y_span) / 2, top)
    bottom_edges = []
    for index, point1 in enumerate(bottom_corners):
        for point2 in bottom_corners[index + 1:]:
            adjacent = sum(coord1 != coord2 for coord1, coord2 in zip(point1, point2)) == 1
            if adjacent:
                bottom_edges.append(Edge(point1, point2))
    assert len(bottom_edges) == 4
    side_edges = [Edge(corner, peak) for corner in bottom_corners]
    assert len(side_edges) == 4
    bottom_surface = PolygonalSurface(tuple(bottom_edges))
    side_surfaces = []
    for bottom_edge in bottom_edges:
        edges = [bottom_edge]
        for side_edge in side_edges:
            if side_edge.start in bottom_edge:
                edges.append(side_edge)
        assert len(edges) == 3
        side_surface = PolygonalSurface(tuple(edges))
        side_surfaces.append(side_surface)
    assert len(side_surfaces) == 4
    return PolygonalShape(tuple([bottom_surface] + side_surfaces))


def make_table(width: float, depth: float) -> PolygonalShape:
    """Construct a PolygonalShape representing the surface of a table (i.e. a
    flat rectangular surface) and return it."""
    left = -width / 2
    right = width / 2
    front = -depth / 2
    back = depth / 2
    bottom = 0
    x_span = (left, right)
    y_span = (front, back)
    corners = []
    for x in x_span:
        for y in y_span:
            corners.append(Point(x, y, bottom))
    assert len(corners) == 4
    edges = []
    for index, point1 in enumerate(corners):
        for point2 in corners[index + 1:]:
            adjacent = sum(coord1 != coord2 for coord1, coord2 in zip(point1, point2)) == 1
            if adjacent:
                edges.append(Edge(point1, point2))
    assert len(edges) == 4, edges
    surface = PolygonalSurface(tuple(edges))
    return PolygonalShape((surface,))


def make_grasper(width: float, height: float) -> PolygonalShape:
    """Construct a PolygonalShape representing a grasper (a mechanical 'hand'
    suspended from a rod or wire) and return it."""
    left = -width / 2
    right = width / 2
    front = -width / 2
    back = width / 2
    top = height
    bottom = 0
    claw_edge1 = Edge(
        Point(left, 0, 0),
        Point(right, 0, 0)
    )
    claw_edge2 = Edge(
        Point(0, front, 0),
        Point(0, back, 0)
    )
    arm_edge = Edge(
        Point(0, 0, bottom),
        Point(0, 0, top)
    )
    return PolygonalShape(
        (PolygonalSurface((claw_edge1,)),
         PolygonalSurface((claw_edge2,)),
         PolygonalSurface((arm_edge,)))
    )


def triangle_area(triangle: Tuple[Point, Point, Point]) -> float:
    """Return the area of a triangle formed by three points in 3D space."""
    a, b, c = triangle
    ab = b - a
    ac = c - a
    return 0.5 * math.sqrt((ab.y * ac.z - ab.z * ac.y) ** 2 +
                           (ab.z * ac.x - ab.x * ac.z) ** 2 +
                           (ab.x * ac.y - ab.y * ac.x) ** 2)


def inside_triangle(point: Point, triangle: Tuple[Point, Point, Point],
                    tolerance: float = 0.001) -> bool:
    """Return whether the given 3D point falls on (or close to) a triangle in
    3D space."""
    abc_area = triangle_area(triangle)
    a, b, c = triangle
    pbc_area = triangle_area((point, b, c))
    apc_area = triangle_area((a, point, c))
    abp_area = triangle_area((a, b, point))
    return abc_area * (1 + tolerance) >= pbc_area + apc_area + abp_area



"""Functionality for viewing and interacting with a scene."""

import textwrap
import threading
import traceback
from typing import Optional, Tuple, Callable, List, Any

import pygame.display

from shrdlu_blocks.geometry import Point
from shrdlu_blocks.control import Controller
from shrdlu_blocks.scenes import Scene, make_standard_scene
from shrdlu_blocks.typedefs import Color

__all__ = ['Viewer']


class Viewer:
    """An interactive scene viewer.

    The viewer provides a fixed-perspective view of the scene. The user can
    type text into the viewer. When the user hits return, the text is submitted
    to a callback, along with a controller associated with the displayed scene.
    The callback can then interact with the scene via the controller based on
    the text entered by the user.
    """

    def __init__(self, screen, title: str = None,
                 callback: Callable[[Controller, str], str] = None):
        self._screen = screen
        self._width = self._screen.get_width()
        self._height = self._screen.get_height()

        self._callback = callback
        self._title = title
        self._zoom = min(self._width, self._height) / 2
        self._screen_center = Point(self._width / 2, 0, self._height / 2)
        self._center_adjustment_ratio = min(self._width, self._height) * 0.01
        self._zoom_adjustment_ratio = 5 / 4
        self._view_center = Point(0, 0, 0)
        self._x_axis = Point(1, 0, 0)
        self._y_axis = Point(0, 1, 0)
        self._z_axis = Point(0, 0, 1)
        self._wall_color = Color(255, 255, 255)
        self._y_to_x_bleed_rate = 0.2
        self._y_to_z_bleed_rate = 0.2

        if title:
            pygame.display.set_caption(title)

        self._scene: Optional[Scene] = make_standard_scene()
        self._controller = Controller(self._scene)

        self._text_height = 25
        self._input_text_color = Color(0, 0, 0)
        self._output_text_color = Color(0, 0, 255)
        self._font = pygame.font.Font(None, self._text_height)
        self._output_text_box = pygame.Rect(0.0, self._height - 2 * self._text_height,
                                            self._width, self._text_height)
        self._input_text_box = pygame.Rect(0.0, self._height - self._text_height,
                                           self._width, self._text_height)
        self._output_text = ''
        self._input_text = ''
        self._input_enabled = True

    @property
    def scene(self) -> Optional[Scene]:
        """The scene displayed by the viewer."""
        return self._scene

    @scene.setter
    def scene(self, scene: Optional[Scene]) -> None:
        """The scene displayed by the viewer."""
        self._scene = scene
        self._controller = None if scene is None else Controller(scene)
        self._input_text = self._output_text = ''

    @property
    def title(self) -> Optional[str]:
        """The title for the viewer window."""
        return self._title

    @title.setter
    def title(self, value: Optional[str]) -> None:
        """The title for the viewer window."""
        if self._title != value:
            pygame.display.set_caption(value or '')
            self._title = value

    @property
    def callback(self) -> Optional[Callable[[Controller, str], str]]:
        """The callback for handling text input to the viewer."""
        return self._callback

    @callback.setter
    def callback(self, value: Optional[Callable[[Controller, str], str]]) -> None:
        """The callback for handling text input to the viewer."""
        self._callback = value

    def move_camera(self, relative_position: Point) -> None:
        self._view_center += relative_position

    def adjust_zoom(self, relative_zoom: float) -> None:
        self._zoom *= relative_zoom

    def run(self):
        alive = True
        while alive:
            try:
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        alive = False
                    elif (self._scene and
                          self._input_enabled and
                          self._callback and
                          event.type == pygame.KEYDOWN):
                        if event.key == pygame.K_RETURN:
                            input_text = self._input_text
                            self._input_text = self._output_text = ''
                            thread = threading.Thread(target=self._run_callback, args=(input_text,))
                            thread.start()
                        elif event.key == pygame.K_BACKSPACE:
                            self._input_text = self._input_text[:-1]
                        else:
                            self._input_text += event.unicode
            except pygame.error as e:
                if 'video system not initialized' in str(e):
                    # Happens sometimes when we quit from another thread.
                    break
                else:
                    raise

            try:
                self.display_scene()
                self.display_input_text_box()
                self.display_output_text_box()
                pygame.display.flip()
            except pygame.error as e:
                if 'display Surface quit' in str(e):
                    # Happens sometimes when we quit from another thread.
                    break
                else:
                    raise

    def display_scene(self) -> None:
        self._screen.fill(self._wall_color)
        if not self._scene:
            return
        polygons = []
        for obj in self._scene.objects:
            transformed_shape = self._screen_center - self._view_center + \
                                (obj.position + obj.shape) * self._zoom
            for surface in transformed_shape.surfaces:
                remaining_edges = set(surface.edges)
                edge = remaining_edges.pop()
                points = [edge.start, edge.end]
                while remaining_edges:
                    for edge in remaining_edges:
                        if edge.start == points[-1]:
                            points.append(edge.end)
                            remaining_edges.remove(edge)
                            break
                        elif edge.end == points[-1]:
                            points.append(edge.start)
                            remaining_edges.remove(edge)
                            break
                    else:
                        assert not remaining_edges, remaining_edges
                polygons.append((obj.color, points))
        polygons.sort(key=lambda polygon: tuple(sorted(((point.y, -point.x, -point.z)
                                                        for point in polygon[1]),
                                                       reverse=True)),
                      reverse=True)
        for color, points in polygons:
            points = [self._map_point_to_screen(point) for point in points]
            if len(points) > 2:
                pygame.draw.polygon(self._screen, color, points)
                pygame.draw.polygon(self._screen, (0, 0, 0), points, width=1)
            else:
                assert len(points) == 2
                pygame.draw.aaline(self._screen, color, points[0], points[1])

    def display_output_text_box(self):
        width, height, surfaces = self._wrap_text(self._output_text, self._output_text_color)
        self._output_text_box.w = width
        self._output_text_box.h = height
        self._output_text_box.top = self._input_text_box.top - height
        for cumulative_height, surface in surfaces:
            self._screen.blit(surface, (self._output_text_box.x + 5,
                                        self._output_text_box.y + 5 + cumulative_height))

    def display_input_text_box(self):
        width, height, surfaces = self._wrap_text(self._input_text, self._input_text_color)
        self._input_text_box.w = width
        self._input_text_box.h = height
        self._input_text_box.top = self._height - height
        self._output_text_box.top = self._input_text_box.top - self._input_text_box.h
        for cumulative_height, surface in surfaces:
            self._screen.blit(surface, (self._input_text_box.x + 5,
                                        self._input_text_box.y + 5 + cumulative_height))

    def _map_point_to_screen(self, point: Point) -> Tuple[float, float]:
        x, y, z = point
        screen_x = x + self._y_to_x_bleed_rate * y
        screen_y = self._height - (z + self._y_to_z_bleed_rate * y)
        return screen_x, screen_y

    def _wrap_text(self, text: str, color: Color) -> Tuple[int, int, List[Tuple[int, Any]]]:
        lines = [line for line in text.splitlines(keepends=False) if line.strip()]
        if not lines:
            lines.append('')
        text_surfaces = [self._font.render(line, True, color) for line in lines]
        max_width = max((surface.get_width() for surface in text_surfaces), default=0) + 10

        # Progressively reduce line width until the wrapped lines fit inside the window.
        line_width = max((len(line) for line in lines), default=0)
        while max_width > self._width:
            line_width -= 1
            text_surfaces = []
            for original_line in lines:
                for line in textwrap.wrap(original_line, width=line_width):
                    text_surfaces.append(self._font.render(line, True, color))
            max_width = max((surface.get_width() for surface in text_surfaces), default=0) + 10

        cumulative_height = [0]
        for surface in text_surfaces:
            cumulative_height.append(cumulative_height[-1] + 5 + surface.get_height())
        total_height = cumulative_height[-1] + 5

        return max_width, total_height, list(zip(cumulative_height, text_surfaces))

    def _run_callback(self, input_text: str) -> None:
        self._input_enabled = False
        for line in input_text.splitlines(keepends=False):
            print('>>> ' + line)
        output_text = ''
        # noinspection PyBroadException
        try:
            assert self._controller is not None
            output_text = self._callback(self._controller, input_text)
            if output_text is None:
                output_text = ''
            elif not isinstance(output_text, str):
                raise TypeError(output_text)
        except Exception:
            output_text = traceback.format_exc()
        finally:
            print(output_text)
            self._output_text = output_text
            self._input_enabled = True
